import time
from typing import List

from paasta_tools.deployd.common import DelayDeadlineQueue
from paasta_tools.deployd.common import PaastaThread
from paasta_tools.deployd.workers import PaastaDeployWorker
from paasta_tools.metrics.metrics_lib import BaseMetrics


class MetricsThread(PaastaThread):
    def __init__(self, metrics_provider):
        super().__init__()
        self.metrics = metrics_provider

    def run_once(self):
        raise NotImplementedError()

    def run(self):
        while True:
            last_run_time = time.time()
            self.run_once()
            time.sleep(last_run_time + 20 - time.time())


class QueueAndWorkerMetrics(MetricsThread):
    def __init__(
        self,
        queue: DelayDeadlineQueue,
        workers: List[PaastaDeployWorker],
        cluster: str,
        metrics_provider: BaseMetrics,
    ) -> None:
        super().__init__(metrics_provider)
        self.daemon = True
        self.queue = queue

        self.instances_to_bounce_later_gauge = self.metrics.create_gauge(
            "instances_to_bounce_later", paasta_cluster=cluster
        )
        self.instances_to_bounce_now_gauge = self.metrics.create_gauge(
            "instances_to_bounce_now", paasta_cluster=cluster
        )
        self.instances_with_past_deadline_gauge = self.metrics.create_gauge(
            "instances_with_past_deadline", paasta_cluster=cluster
        )
        self.instances_with_deadline_in_next_n_seconds_gauges = {
            (available, n): self.metrics.create_gauge(
                f"{available}_instances_with_deadline_in_next_{n}s",
                paasta_cluster=cluster,
            )
            for n in [60, 300, 3600]
            for available in ["available", "unavailable"]
        }
        self.max_time_past_deadline_gauge = self.metrics.create_gauge(
            "max_time_past_deadline", paasta_cluster=cluster
        )
        self.sum_time_past_deadline_gauge = self.metrics.create_gauge(
            "sum_time_past_deadline", paasta_cluster=cluster
        )

        self.workers = workers

        self.workers_busy_gauge = self.metrics.create_gauge(
            "workers_busy", paasta_cluster=cluster
        )
        self.workers_idle_gauge = self.metrics.create_gauge(
            "workers_idle", paasta_cluster=cluster
        )
        self.workers_dead_gauge = self.metrics.create_gauge(
            "workers_dead", paasta_cluster=cluster
        )

    def run_once(self) -> None:
        self.instances_to_bounce_later_gauge.set(
            self.queue.unavailable_service_instances.qsize()
        )
        self.instances_to_bounce_now_gauge.set(
            self.queue.available_service_instances.qsize()
        )

        currently_available_instances = tuple(
            self.queue.available_service_instances.queue
        )
        currently_unavailable_instances = tuple(
            self.queue.unavailable_service_instances.queue
        )

        available_deadlines = [
            deadline for deadline, _ in currently_available_instances
        ]
        unavailable_deadlines = [
            deadline for _, deadline, _ in currently_unavailable_instances
        ]

        now = time.time()
        self.instances_with_past_deadline_gauge.set(
            len([1 for deadline in available_deadlines if deadline < now])
        )

        for (
            (available, n),
            gauge,
        ) in self.instances_with_deadline_in_next_n_seconds_gauges.items():
            if available == "available":
                deadlines = available_deadlines
            else:
                deadlines = unavailable_deadlines

            gauge.set(len([1 for deadline in deadlines if now < deadline < now + n]))

        self.max_time_past_deadline_gauge.set(
            max(
                [now - deadline for deadline in available_deadlines if deadline < now],
                default=0,
            )
        )

        self.sum_time_past_deadline_gauge.set(
            sum([max(0, now - deadline) for deadline in available_deadlines])
        )

        self.workers_busy_gauge.set(
            len([worker for worker in self.workers if worker.busy])
        )
        self.workers_idle_gauge.set(
            len([worker for worker in self.workers if not worker.busy])
        )
        self.workers_dead_gauge.set(
            len([worker for worker in self.workers if not worker.is_alive()])
        )



# Copyright 2015-2018 Yelp Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from urllib.parse import urljoin

import requests
import yaml

from paasta_tools.utils import get_user_agent


log = logging.getLogger(__name__)


class TronRequestError(Exception):
    pass


class TronClient:
    """
    Client for interacting with a Tron master.
    """

    def __init__(self, url):
        self.master_url = url

    def _request(self, method, url, data):
        headers = {"User-Agent": get_user_agent()}
        kwargs = {"url": urljoin(self.master_url, url), "headers": headers}
        if method == "GET":
            kwargs["params"] = data
            response = requests.get(**kwargs)
        elif method == "POST":
            kwargs["data"] = data
            response = requests.post(**kwargs)
        else:
            raise ValueError(f"Unrecognized method: {method}")

        return self._get_response_or_error(response)

    def _get_response_or_error(self, response):
        try:
            result = response.json()
            if "error" in result:
                raise TronRequestError(result["error"])
            return result
        except ValueError:  # Not JSON
            if not response.ok:
                raise TronRequestError(
                    "Status code {status_code} for {url}: {reason}".format(
                        status_code=response.status_code,
                        url=response.url,
                        reason=response.reason,
                    )
                )
            return response.text

    def _get(self, url, data=None):
        return self._request("GET", url, data)

    def _post(self, url, data=None):
        return self._request("POST", url, data)

    def update_namespace(self, namespace, new_config, skip_if_unchanged=True):
        """Updates the configuration for a namespace.

        :param namespace: str
        :param new_config: str, should be valid YAML.
        :param skip_if_unchanged: boolean. If False, will send the update
            even if the current config matches the new config.
        """
        current_config = self._get("/api/config", {"name": namespace, "no_header": 1})

        if skip_if_unchanged:
            if yaml.safe_load(new_config) == yaml.safe_load(current_config["config"]):
                log.debug("No change in config, skipping update.")
                return

        return self._post(
            "/api/config",
            data={
                "name": namespace,
                "config": new_config,
                "hash": current_config["hash"],
                "check": 0,
            },
        )

    def list_namespaces(self):
        """Gets the namespaces that are currently configured."""
        response = self._get("/api")
        return response.get("namespaces", [])

    def get_job_content(self, job: str) -> dict:
        return self._get(f"/api/jobs/{job}/")

    def get_latest_job_run_id(self, job_content: dict) -> str:
        job_runs = sorted(
            job_content.get("runs", []),
            key=lambda k: (k["state"] != "scheduled", k["run_num"]),
            reverse=True,
        )
        return job_runs[0]["run_num"]

    def get_action_run(self, job: str, action: str, run_id: str) -> dict:
        return self._get(
            f"/api/jobs/{job}/{run_id}/{action}?include_stderr=1&include_stdout=1&num_lines=10"
        )



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import datetime
import itertools
import json
import logging
import re
import socket
from collections import namedtuple
from pathlib import Path
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Collection
from typing import Dict
from typing import List
from typing import Mapping
from typing import MutableMapping
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from urllib.parse import urlparse

import a_sync
import humanize
import requests
from kazoo.client import KazooClient
from mypy_extensions import TypedDict

import paasta_tools.mesos.cluster as cluster
import paasta_tools.mesos.exceptions as mesos_exceptions
from paasta_tools.async_utils import aiter_to_list
from paasta_tools.async_utils import async_timeout
from paasta_tools.async_utils import async_ttl_cache
from paasta_tools.long_running_service_tools import host_passes_blacklist
from paasta_tools.long_running_service_tools import host_passes_whitelist
from paasta_tools.mesos.cfg import load_mesos_config
from paasta_tools.mesos.exceptions import SlaveDoesNotExist
from paasta_tools.mesos.master import MesosMaster
from paasta_tools.mesos.master import MesosState
from paasta_tools.mesos.task import Task
from paasta_tools.utils import DeployBlacklist
from paasta_tools.utils import DeployWhitelist
from paasta_tools.utils import format_table
from paasta_tools.utils import get_user_agent
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import TimeoutError

MARATHON_FRAMEWORK_NAME_PREFIX = "marathon"

ZookeeperHostPath = namedtuple("ZookeeperHostPath", ["host", "path"])
SlaveTaskCount = namedtuple("SlaveTaskCount", ["count", "batch_count", "slave"])

DEFAULT_MESOS_CLI_CONFIG_LOCATION = "/nail/etc/mesos-cli.json"

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def get_mesos_config_path():
    """
    Determine where to find the configuration for mesos-cli.
    """
    return (
        load_system_paasta_config()
        .get_mesos_cli_config()
        .get("path", DEFAULT_MESOS_CLI_CONFIG_LOCATION)
    )


def get_mesos_config():
    return load_mesos_config(get_mesos_config_path())


def get_mesos_master(**overrides: Any) -> MesosMaster:
    config = get_mesos_config()
    for k, v in overrides.items():
        config[k] = v
    return MesosMaster(config)


MY_HOSTNAME = socket.getfqdn()
MESOS_MASTER_PORT = 5050
MESOS_SLAVE_PORT = "5051"


class MesosSlaveConnectionError(Exception):
    pass


class MesosTailLines(NamedTuple):
    stdout: List[str]
    stderr: List[str]
    error_message: str


def get_mesos_leader() -> str:
    """Get the current mesos-master leader's hostname.
    Attempts to determine this by using mesos.cli to query ZooKeeper.

    :returns: The current mesos-master hostname"""
    try:
        url = get_mesos_master().host
    except mesos_exceptions.MasterNotAvailableException:
        log.debug("mesos.cli failed to provide the master host")
        raise
    log.debug("mesos.cli thinks the master host is: %s" % url)
    hostname = urlparse(url).hostname
    log.debug("The parsed master hostname is: %s" % hostname)
    # This check is necessary, as if we parse a value such as 'localhost:5050',
    # it won't have a hostname attribute
    if hostname:
        try:
            host = socket.gethostbyaddr(hostname)[0]
            fqdn = socket.getfqdn(host)
        except (socket.error, socket.herror, socket.gaierror, socket.timeout):
            log.debug("Failed to convert mesos leader hostname to fqdn!")
            raise
        log.debug("Mesos Leader: %s" % fqdn)
        return fqdn
    else:
        raise ValueError("Expected to receive a valid URL, got: %s" % url)


def is_mesos_leader(hostname: str = MY_HOSTNAME) -> bool:
    """Check if a hostname is the current mesos leader.

    :param hostname: The hostname to query mesos-master on
    :returns: True if hostname is the mesos-master leader, False otherwise"""
    return get_mesos_leader() == hostname


def find_mesos_leader(master):
    """ Find the leader with redirect given one mesos master.
    """
    if master is None:
        raise ValueError("Mesos master is required to find leader")

    url = f"http://{master}:{MESOS_MASTER_PORT}/redirect"
    response = requests.get(url)
    return urlparse(response.url).hostname


async def get_current_tasks(job_id: str) -> List[Task]:
    """ Returns a list of all the tasks with a given job id.
    :param job_id: the job id of the tasks.
    :return tasks: a list of mesos.cli.Task.
    """
    mesos_master = get_mesos_master()
    framework_tasks = await mesos_master.tasks(fltr=job_id, active_only=False)
    return framework_tasks


def is_task_running(task: Task) -> bool:
    return task["state"] == "TASK_RUNNING"


def filter_running_tasks(tasks: Collection[Task]) -> List[Task]:
    """ Filters those tasks where it's state is TASK_RUNNING.
    :param tasks: a list of mesos.cli.Task
    :return filtered: a list of running tasks
    """
    return [task for task in tasks if is_task_running(task)]


def filter_not_running_tasks(tasks: Collection[Task]) -> List[Task]:
    """ Filters those tasks where it's state is *not* TASK_RUNNING.
    :param tasks: a list of mesos.cli.Task
    :return filtered: a list of tasks *not* running
    """
    return [task for task in tasks if not is_task_running(task)]


async def get_running_tasks_from_frameworks(job_id=""):
    """ Will include tasks from active and completed frameworks
    but NOT orphaned tasks
    """
    active_framework_tasks = await get_current_tasks(job_id)
    running_tasks = filter_running_tasks(active_framework_tasks)
    return running_tasks


async def get_all_running_tasks() -> Collection[Task]:
    """ Will include all running tasks; for now orphans are not included
    """
    framework_tasks = await get_current_tasks("")
    mesos_master = get_mesos_master()
    framework_tasks += await mesos_master.orphan_tasks()
    running_tasks = filter_running_tasks(framework_tasks)
    return running_tasks


@async_ttl_cache(ttl=600)
async def get_cached_list_of_all_current_tasks():
    """Returns a cached list of all mesos tasks.

    This function is used by 'paasta status' and 'paasta_serviceinit status'
    to avoid re-querying mesos master and re-parsing json to get mesos.Task objects.


    The async_ttl_cache decorator caches the list for 600 seconds.
    ttl doesn't really matter for this function because when we run 'paasta status'
    the corresponding HTTP request to mesos master is cached by requests_cache.

    :return tasks: a list of mesos.Task
    """
    return await get_current_tasks("")


@async_ttl_cache(ttl=600)
async def get_cached_list_of_running_tasks_from_frameworks():
    """Returns a cached list of all running mesos tasks.
    See the docstring for get_cached_list_of_all_current_tasks().

    :return tasks: a list of mesos.Task
    """
    return [
        task
        for task in filter_running_tasks(await get_cached_list_of_all_current_tasks())
    ]


@async_ttl_cache(ttl=600)
async def get_cached_list_of_not_running_tasks_from_frameworks():
    """Returns a cached list of mesos tasks that are NOT running.
    See the docstring for get_cached_list_of_all_current_tasks().

    :return tasks: a list of mesos.Task"""
    return [
        task
        for task in filter_not_running_tasks(
            await get_cached_list_of_all_current_tasks()
        )
    ]


def select_tasks_by_id(tasks: Collection[Task], job_id: str = "") -> List[Task]:
    """Returns a list of the tasks with a given job_id.

    :param tasks: a list of mesos.Task.
    :param job_id: the job id.
    :return tasks: a list of mesos.Task.
    """
    return [task for task in tasks if job_id in task["id"]]


async def get_non_running_tasks_from_frameworks(job_id: str = "") -> List[Task]:
    """ Will include tasks from active and completed frameworks
    but NOT orphaned tasks
    """
    active_framework_tasks = await get_current_tasks(job_id)
    not_running_tasks = filter_not_running_tasks(active_framework_tasks)
    return not_running_tasks


async def get_short_hostname_from_task(task: Task) -> str:
    try:
        slave_hostname = (await task.slave())["hostname"]
        return slave_hostname.split(".")[0]
    except (AttributeError, SlaveDoesNotExist):
        return "Unknown"


def get_first_status_timestamp(task: Task) -> Optional[float]:
    try:
        start_time_string = task["statuses"][0]["timestamp"]
        return float(start_time_string)
    except (IndexError, SlaveDoesNotExist):
        return None


def get_first_status_timestamp_string(task: Task) -> str:
    """Gets the first status timestamp from a task id and returns a human
    readable string with the local time and a humanized duration:
    ``2015-01-30T08:45 (an hour ago)``
    """
    first_status_timestamp = get_first_status_timestamp(task)
    if first_status_timestamp is None:
        return "Unknown"
    else:
        first_status_datetime = datetime.datetime.fromtimestamp(first_status_timestamp)
        return "{} ({})".format(
            first_status_datetime.strftime("%Y-%m-%dT%H:%M"),
            humanize.naturaltime(first_status_datetime),
        )


async def get_mem_usage(task: Task) -> str:
    try:
        task_mem_limit = await task.mem_limit()
        task_rss = await task.rss()
        if task_mem_limit == 0:
            return "Undef"
        mem_percent = task_rss / task_mem_limit * 100
        mem_string = "%d/%dMB" % (
            (task_rss / 1024 / 1024),
            (task_mem_limit / 1024 / 1024),
        )
        if mem_percent > 90:
            return PaastaColors.red(mem_string)
        else:
            return mem_string
    except (AttributeError, SlaveDoesNotExist):
        return "None"
    except TimeoutError:
        return "Timed Out"


async def get_cpu_shares(task: Task) -> float:
    # The CPU shares has an additional .1 allocated to it for executor overhead.
    # We subtract this to the true number
    # (https://github.com/apache/mesos/blob/dc7c4b6d0bcf778cc0cad57bb108564be734143a/src/slave/constants.hpp#L100)
    cpu_shares = await task.cpu_limit()
    return cpu_shares - 0.1


async def get_cpu_usage(task: Task) -> str:
    """Calculates a metric of used_cpu/allocated_cpu
    To do this, we take the total number of cpu-seconds the task has consumed,
    (the sum of system and user time), OVER the total cpu time the task
    has been allocated.

    The total time a task has been allocated is the total time the task has
    been running (https://github.com/mesosphere/mesos/blob/0b092b1b0/src/webui/master/static/js/controllers.js#L140)
    multiplied by the "shares" a task has.
    """
    try:
        start_time = round(task["statuses"][0]["timestamp"])
        current_time = int(datetime.datetime.now().strftime("%s"))
        duration_seconds = current_time - start_time
        cpu_shares = await get_cpu_shares(task)
        allocated_seconds = duration_seconds * cpu_shares
        task_stats = await task.stats()
        used_seconds = task_stats.get("cpus_system_time_secs", 0.0) + task_stats.get(
            "cpus_user_time_secs", 0.0
        )
        if allocated_seconds == 0:
            return "Undef"
        percent = round(100 * (used_seconds / allocated_seconds), 1)
        percent_string = "%s%%" % percent
        if percent > 90:
            return PaastaColors.red(percent_string)
        else:
            return percent_string
    except (AttributeError, SlaveDoesNotExist):
        return "None"
    except TimeoutError:
        return "Timed Out"


async def results_or_unknown(future: Awaitable[str]) -> str:
    try:
        return await future
    except Exception:
        return PaastaColors.red("Unknown")


async def format_running_mesos_task_row(
    task: Task, get_short_task_id: Callable[[str], str]
) -> Tuple[str, ...]:
    """Returns a pretty formatted string of a running mesos task attributes"""

    short_task_id = get_short_task_id(task["id"])
    short_hostname_future = asyncio.ensure_future(
        results_or_unknown(get_short_hostname_from_task(task))
    )
    mem_usage_future = asyncio.ensure_future(results_or_unknown(get_mem_usage(task)))
    cpu_usage_future = asyncio.ensure_future(results_or_unknown(get_cpu_usage(task)))
    first_status_timestamp = get_first_status_timestamp_string(task)

    await asyncio.wait([short_hostname_future, mem_usage_future, cpu_usage_future])

    return (
        short_task_id,
        short_hostname_future.result(),
        mem_usage_future.result(),
        cpu_usage_future.result(),
        first_status_timestamp,
    )


async def format_non_running_mesos_task_row(
    task: Task, get_short_task_id: Callable[[str], str]
) -> Tuple[str, ...]:
    """Returns a pretty formatted string of a running mesos task attributes"""
    return (
        PaastaColors.grey(get_short_task_id(task["id"])),
        PaastaColors.grey(await results_or_unknown(get_short_hostname_from_task(task))),
        PaastaColors.grey(get_first_status_timestamp_string(task)),
        PaastaColors.grey(task["state"]),
    )


@async_timeout()
async def get_tail_lines_for_mesos_task(
    task: Task, get_short_task_id: Callable[[str], str], num_tail_lines: int
) -> MutableMapping[str, Sequence[str]]:
    tail_lines_dict: MutableMapping[str, Sequence[str]] = {}
    mesos_cli_config = get_mesos_config()

    try:
        fobjs = await aiter_to_list(
            cluster.get_files_for_tasks(
                task_list=[task],
                file_list=["stdout", "stderr"],
                max_workers=mesos_cli_config["max_workers"],
            )
        )
        if not fobjs:
            return {"stdout": [], "stderr": []}

        fobjs.sort(key=lambda fobj: fobj.path, reverse=True)

        for fobj in fobjs:
            # read nlines, starting from EOF
            tail = []
            lines_seen = 0

            async for line in fobj._readlines_reverse():
                tail.append(line)
                lines_seen += 1
                if lines_seen >= num_tail_lines:
                    break

            # reverse the tail, so that EOF is at the bottom again
            tail_lines_dict[fobj.path] = tail[::-1]
    except (
        mesos_exceptions.MasterNotAvailableException,
        mesos_exceptions.SlaveDoesNotExist,
        mesos_exceptions.TaskNotFoundException,
        mesos_exceptions.FileNotFoundForTaskException,
        TimeoutError,
    ) as e:
        short_task_id = get_short_task_id(task["id"])
        error_name = e.__class__.__name__
        return {
            "error_message": f"couldn't read stdout/stderr for {short_task_id} ({error_name})"
        }

    return tail_lines_dict


def format_tail_lines_for_mesos_task(tail_lines, task_id):
    rows = []
    if (tail_lines.stderr or tail_lines.stdout) is not None:
        if len(tail_lines.stderr) + len(tail_lines.stdout) == 0:
            rows.append(PaastaColors.blue(f"  no stdout/stderrr for {task_id}"))
        else:
            for stdstream in ("stdout", "stderr"):
                rows.append(PaastaColors.blue(f"{stdstream} tail for {task_id}"))
                rows.extend(f"  {line}" for line in getattr(tail_lines, stdstream, []))
    elif tail_lines.error_message is not None:
        rows.append(PaastaColors.red(f"  {tail_lines.error_message}"))

    return rows


@async_timeout()
async def format_stdstreams_tail_for_task(task, get_short_task_id, nlines=10):
    tail_lines_dict = await get_tail_lines_for_mesos_task(
        task, get_short_task_id, nlines
    )
    tail_lines = MesosTailLines(
        stdout=tail_lines_dict.get("stdout"),
        stderr=tail_lines_dict.get("stderr"),
        error_message=tail_lines_dict.get("error_message"),
    )
    return [
        f"    {line}"
        for line in format_tail_lines_for_mesos_task(tail_lines, task["id"])
    ]


def zip_tasks_verbose_output(table, stdstreams):
    """Zip a list of strings (table) with a list of lists (stdstreams)
    :param table: a formatted list of tasks
    :param stdstreams: for each task, a list of lines from stdout/stderr tail
    """
    if len(table) != len(stdstreams):
        raise ValueError("Can only zip same-length lists")
    output = []
    for i in range(len(table)):
        output.append(table[i])
        output.extend([line for line in stdstreams[i]])
    return output


async def format_task_list(
    tasks: Sequence[Task],
    list_title: str,
    table_header: Sequence[str],
    get_short_task_id: Callable[[str], str],
    format_task_row: Callable[
        [Task, Callable[[str], str]], Awaitable[Union[Sequence[str], str]]
    ],
    grey: bool,
    tail_lines: int,
) -> List[str]:
    """Formats a list of tasks, returns a list of output lines
    :param tasks: List of tasks as returned by get_*_tasks_from_all_frameworks.
    :param list_title: 'Running Tasks:' or 'Non-Running Tasks'.
    :param table_header: List of column names used in the tasks table.
    :param get_short_task_id: A function which given a task_id returns a short task_id suitable for printing.
    :param format_task_row: Formatting function, works on a task and a get_short_task_id function.
    :param tail_lines (int): number of lines of stdout/stderr to tail, as obtained from the Mesos sandbox.
    :param grey: If True, the list will be made less visually prominent.
    :return output: Formatted output (list of output lines).
    """
    if not grey:

        def colorize(x):
            return x

    else:

        def colorize(x):
            return PaastaColors.grey(x)

    output = []
    output.append(colorize("  %s" % list_title))
    table_rows: List[Union[str, Sequence[str]]] = [
        [colorize(th) for th in table_header]
    ]

    if tasks:
        task_row_futures = [
            asyncio.ensure_future(format_task_row(task, get_short_task_id))
            for task in tasks
        ]
        await asyncio.wait(task_row_futures)

        for future in task_row_futures:
            table_rows.append(future.result())

    tasks_table = ["    %s" % row for row in format_table(table_rows)]
    if tail_lines == 0:
        output.extend(tasks_table)
    else:
        stdstreams = []
        for task in tasks:
            stdstreams.append(
                await format_stdstreams_tail_for_task(
                    task, get_short_task_id, nlines=tail_lines
                )
            )
        output.append(tasks_table[0])  # header
        output.extend(zip_tasks_verbose_output(tasks_table[1:], stdstreams))

    return output


@a_sync.to_blocking
async def status_mesos_tasks_verbose(
    filter_string: str, get_short_task_id: Callable[[str], str], tail_lines: int = 0
) -> str:
    """Returns detailed information about the mesos tasks for a service.

    :param filter_string: An id used for looking up Mesos tasks
    :param get_short_task_id: A function which given a
                              task_id returns a short task_id suitable for
                              printing.
    :param tail_lines: int representing the number of lines of stdout/err to
                       report.
    """
    output: List[str] = []
    running_and_active_tasks = select_tasks_by_id(
        await get_cached_list_of_running_tasks_from_frameworks(), filter_string
    )
    list_title = "Running Tasks:"
    table_header = [
        "Mesos Task ID",
        "Host deployed to",
        "Ram",
        "CPU",
        "Deployed at what localtime",
    ]
    output.extend(
        await format_task_list(
            tasks=running_and_active_tasks,
            list_title=list_title,
            table_header=table_header,
            get_short_task_id=get_short_task_id,
            format_task_row=format_running_mesos_task_row,
            grey=False,
            tail_lines=tail_lines,
        )
    )

    non_running_tasks = select_tasks_by_id(
        await get_cached_list_of_not_running_tasks_from_frameworks(), filter_string
    )
    # Order the tasks by timestamp
    non_running_tasks.sort(key=lambda task: get_first_status_timestamp_string(task))
    non_running_tasks_ordered = list(reversed(non_running_tasks[-10:]))

    list_title = "Non-Running Tasks"
    table_header = [
        "Mesos Task ID",
        "Host deployed to",
        "Deployed at what localtime",
        "Status",
    ]
    output.extend(
        await format_task_list(
            tasks=non_running_tasks_ordered,
            list_title=list_title,
            table_header=table_header,
            get_short_task_id=get_short_task_id,
            format_task_row=format_non_running_mesos_task_row,
            grey=True,
            tail_lines=tail_lines,
        )
    )

    return "\n".join(output)


def get_local_slave_state(hostname=None):
    """Fetches mesos slave state and returns it as a dict.

    :param hostname: The host from which to fetch slave state. If not specified, defaults to the local machine."""
    if hostname is None:
        hostname = socket.getfqdn()
    stats_uri = f"http://{hostname}:{MESOS_SLAVE_PORT}/state"
    try:
        headers = {"User-Agent": get_user_agent()}
        response = requests.get(stats_uri, timeout=10, headers=headers)
        if response.status_code == 404:
            fallback_stats_uri = f"http://{hostname}:{MESOS_SLAVE_PORT}/state.json"
            response = requests.get(fallback_stats_uri, timeout=10, headers=headers)
    except requests.ConnectionError as e:
        raise MesosSlaveConnectionError(
            "Could not connect to the mesos slave to see which services are running\n"
            "on %s. Is the mesos-slave running?\n"
            "Error was: %s\n" % (e.request.url, str(e))
        )
    response.raise_for_status()
    return json.loads(response.text)


async def get_mesos_quorum():
    """Returns the configured quorum size."""
    return int((await get_master_flags())["flags"]["quorum"])


MesosResources = Mapping[str, Any]


class MesosTask(TypedDict):
    resources: MesosResources
    slave_id: str
    id: str
    state: str


def get_all_tasks_from_state(
    mesos_state: MesosState, include_orphans: bool = False
) -> Sequence[MesosTask]:
    """Given a mesos state, find the tasks from all frameworks.
    :param mesos_state: the mesos_state
    :returns: a list of tasks
    """
    tasks = [
        task
        for framework in mesos_state.get("frameworks", [])
        for task in framework.get("tasks", [])
    ]
    if include_orphans:
        tasks += mesos_state.get("orphan_tasks", [])
    return tasks


async def get_master_flags():
    res = await get_mesos_master().fetch("/master/flags")
    return await res.json()


def get_zookeeper_host_path():
    zk_url = "zk://%s" % load_system_paasta_config().get_zk_hosts()
    parsed = urlparse(zk_url)
    return ZookeeperHostPath(host=parsed.netloc, path=parsed.path)


def get_zookeeper_config(state):
    """Returns dict, containing the zookeeper hosts and path.
    :param state: mesos state dictionary"""
    re_zk = re.match(r"^zk://([^/]*)/(.*)$", state["flags"]["zk"])
    return {"hosts": re_zk.group(1), "path": re_zk.group(2)}


def get_number_of_mesos_masters(host, path):
    """Returns an array, containing mesos masters
    :param zk_config: dict containing information about zookeeper config.
    Masters register themselves in zookeeper by creating ``info_`` entries.
    We count these entries to get the number of masters.
    """
    zk = KazooClient(hosts=host, read_only=True)
    zk.start()
    try:
        root_entries = zk.get_children(path)
        result = [
            info
            for info in root_entries
            if info.startswith("json.info_") or info.startswith("info_")
        ]
        return len(result)
    finally:
        zk.stop()
        zk.close()


def get_all_slaves_for_blacklist_whitelist(
    blacklist: DeployBlacklist, whitelist: DeployWhitelist
):
    """
    A wrapper function to get all slaves and filter according to
    provided blacklist and whitelist.

    :param blacklist: a blacklist, used to filter mesos slaves by attribute
    :param whitelist: a whitelist, used to filter mesos slaves by attribute

    :returns: a list of mesos slave objects, filtered by those which are acceptable
    according to the provided blacklist and whitelists.
    """
    all_slaves = get_slaves()
    return filter_mesos_slaves_by_blacklist(all_slaves, blacklist, whitelist)


def get_mesos_slaves_grouped_by_attribute(slaves, attribute):
    """Returns a dictionary of unique values and the corresponding hosts for a given Mesos attribute

    :param slaves: a list of mesos slaves to group
    :param attribute: an attribute to filter
    :returns: a dictionary of the form {'<attribute_value>': [<list of hosts with attribute=attribute_value>]}
              (response can contain multiple 'attribute_value)
    """
    sorted_slaves = sorted(slaves, key=lambda slave: slave["attributes"].get(attribute))
    return {
        key: list(group)
        for key, group in itertools.groupby(
            sorted_slaves, key=lambda slave: slave["attributes"].get(attribute)
        )
        if key
    }


# TODO: remove to_blocking, convert call sites (smartstack_tools and marathon_serviceinit) to asyncio.
@a_sync.to_blocking
async def get_slaves():
    return (await (await get_mesos_master().fetch("/master/slaves")).json())["slaves"]


def filter_mesos_slaves_by_blacklist(
    slaves, blacklist: DeployBlacklist, whitelist: DeployWhitelist
):
    """Takes an input list of slaves and filters them based on the given blacklist.
    The blacklist is in the form of:

        [["location_type", "location]]

    Where the list inside is something like ["region", "uswest1-prod"]

    :returns: The list of mesos slaves after the filter
    """
    filtered_slaves = []
    for slave in slaves:
        if host_passes_blacklist(
            slave["attributes"], blacklist
        ) and host_passes_whitelist(slave["attributes"], whitelist):
            filtered_slaves.append(slave)
    return filtered_slaves


def get_container_id_for_mesos_id(client, mesos_task_id):
    running_containers = client.containers()

    container_id = None
    for container in running_containers:
        info = client.inspect_container(container)
        if info["Config"]["Env"]:
            for env_var in info["Config"]["Env"]:
                if ("MESOS_TASK_ID=%s" % mesos_task_id) in env_var:
                    container_id = info["Id"]
                    break

    return container_id


def get_mesos_id_from_container(container, client):
    mesos_id = None
    info = client.inspect_container(container)
    if info["Config"]["Env"]:
        for env_var in info["Config"]["Env"]:
            # In marathon it is like this
            if "MESOS_TASK_ID=" in env_var:
                mesos_id = re.match("MESOS_TASK_ID=(.*)", env_var).group(1)
                break
            # Chronos it is like this?
            if "mesos_task_id=" in env_var:
                mesos_id = re.match("mesos_task_id=(.*)", env_var).group(1)
                break
    return mesos_id


def get_mesos_network_for_net(net):
    docker_mesos_net_mapping = {"none": "NONE", "bridge": "BRIDGE", "host": "HOST"}
    return docker_mesos_net_mapping.get(net, net)


async def get_mesos_task_count_by_slave(
    mesos_state: MesosState,
    slaves_list: Sequence[Dict] = None,
    pool: Optional[str] = None,
) -> List[Dict]:
    """Get counts of running tasks per mesos slave. Also include separate count of batch tasks

    :param mesos_state: mesos state dict
    :param slaves_list: a list of slave dicts to count running tasks for.
    :param pool: pool of slaves to return (None means all)
    :returns: list of slave dicts {'task_count': SlaveTaskCount}
    """
    all_mesos_tasks = await get_all_running_tasks()  # empty string = all app ids
    slaves = {
        slave["id"]: {"count": 0, "slave": slave, "batch_count": 0}
        for slave in mesos_state.get("slaves", [])
    }
    for task in all_mesos_tasks:
        try:
            task_slave = await task.slave()
            if task_slave["id"] not in slaves:
                log.debug("Slave {} not found for task".format(task_slave["id"]))
                continue
            else:
                slaves[task_slave["id"]]["count"] += 1
                task_framework = await task.framework()
                log.debug(f"Task framework: {task_framework.name}")
                # Marathon is only framework that runs service. Others are batch.
                if not task_framework.name.startswith(MARATHON_FRAMEWORK_NAME_PREFIX):
                    slaves[task_slave["id"]]["batch_count"] += 1
        except SlaveDoesNotExist:
            log.debug(
                "Tried to get mesos slaves for task {}, but none existed.".format(
                    task["id"]
                )
            )
            continue
    if slaves_list:
        for slave in slaves_list:
            slave["task_counts"] = SlaveTaskCount(
                **slaves[slave["task_counts"].slave["id"]]
            )
        slaves_with_counts = list(slaves_list)
    elif pool:
        slaves_with_counts = [
            {"task_counts": SlaveTaskCount(**slave_counts)}
            for slave_counts in slaves.values()
            if slave_counts["slave"]["attributes"].get("pool", "default") == pool
        ]
    else:
        slaves_with_counts = [
            {"task_counts": SlaveTaskCount(**slave_counts)}
            for slave_counts in slaves.values()
        ]
    for slave in slaves_with_counts:
        log.debug(
            "Slave: {}, running {} tasks, "
            "including {} batch tasks".format(
                slave["task_counts"].slave["hostname"],
                slave["task_counts"].count,
                slave["task_counts"].batch_count,
            )
        )
    return slaves_with_counts


def get_count_running_tasks_on_slave(hostname: str) -> int:
    """Return the number of tasks running on a particular slave
    or 0 if the slave is not found.
    :param hostname: hostname of the slave
    :returns: integer count of mesos tasks"""
    mesos_state = a_sync.block(get_mesos_master().state_summary)
    task_counts = a_sync.block(get_mesos_task_count_by_slave, mesos_state)
    counts = [
        slave["task_counts"].count
        for slave in task_counts
        if slave["task_counts"].slave["hostname"] == hostname
    ]
    if counts:
        return counts[0]
    else:
        return 0


def slave_pid_to_ip(slave_pid: str) -> str:
    """Convert slave_pid to IP

    :param: slave pid e.g. slave(1)@10.40.31.172:5051
    :returns: ip address"""
    regex = re.compile(r".+?@([\d\.]+):\d+")
    return regex.match(slave_pid).group(1)


async def list_framework_ids(active_only=False):
    return [f.id for f in await get_mesos_master().frameworks(active_only=active_only)]


@a_sync.to_blocking
async def get_all_frameworks(active_only=False):
    return await get_mesos_master().frameworks(active_only=active_only)


def terminate_framework(framework_id):
    resp = requests.post(
        "http://%s:%d/master/teardown" % (get_mesos_leader(), MESOS_MASTER_PORT),
        data={"frameworkId": framework_id},
    )
    resp.raise_for_status()


async def get_tasks_from_app_id(app_id, slave_hostname=None):
    tasks = await get_running_tasks_from_frameworks(app_id)
    if slave_hostname:
        tasks = [
            task
            for task in tasks
            if await filter_task_by_hostname(task, slave_hostname)
        ]
    return tasks


async def get_task(task_id: str, app_id: str = "") -> MesosTask:
    tasks = await get_running_tasks_from_frameworks(app_id)
    tasks = [task for task in tasks if filter_task_by_task_id(task, task_id)]
    if len(tasks) < 1:
        raise TaskNotFound(f"Couldn't find task for given id: {task_id}")
    if len(tasks) > 1:
        raise TooManyTasks(
            f"Found more than one task with id: {task_id}, this should not happen!"
        )
    return tasks[0]


def filter_task_by_task_id(task: MesosTask, task_id: str) -> bool:
    return task["id"] == task_id


async def filter_task_by_hostname(task, hostname):
    return (await task.slave())["hostname"].startswith(hostname)


class TaskNotFound(Exception):
    pass


class TooManyTasks(Exception):
    pass


# TODO: async this
def mesos_services_running_here(
    framework_filter, parse_service_instance_from_executor_id, hostname=None
):
    """See what paasta_native services are being run by a mesos-slave on this host.

    :param framework_filter: a function that returns true if we should consider a given framework.
    :param parse_service_instance_from_executor_id: A function that returns a tuple of (service, instance) from the
                                                    executor ID.
    :param hostname: Hostname to fetch mesos slave state from. See get_local_slave_state.

    :returns: A list of triples of (service, instance, port)"""
    slave_state = get_local_slave_state(hostname=hostname)
    frameworks = [
        fw for fw in slave_state.get("frameworks", []) if framework_filter(fw)
    ]
    executors = [
        ex
        for fw in frameworks
        for ex in fw.get("executors", [])
        if "TASK_RUNNING" in [t["state"] for t in ex.get("tasks", [])]
    ]
    srv_list = []
    for executor in executors:
        try:
            srv_name, srv_instance = parse_service_instance_from_executor_id(
                executor["id"]
            )
        except ValueError:
            log.error(
                "Failed to decode paasta service instance from {}".format(
                    executor["id"]
                )
            )
            continue
        if "ports" in executor["resources"]:
            srv_port = int(re.findall("[0-9]+", executor["resources"]["ports"])[0])
        else:
            srv_port = None
        srv_list.append((srv_name, srv_instance, srv_port))
    return srv_list


def is_task_terminal(task: MesosTask,) -> bool:
    """Return whether a given mesos task is terminal.

    Terminal states are documented in
    http://mesos.apache.org/api/latest/java/org/apache/mesos/Protos.TaskState.html

    :param task: the task to be inspected
    :returns: a boolean indicating if the task is considered to be in a terminal state
    """
    return task["state"] in [
        "TASK_ERROR",
        "TASK_KILLED",
        "TASK_FAILED",
        "TASK_FINISHED",
    ]


def is_mesos_available() -> bool:
    return Path(get_mesos_config_path()).exists()



#!/usr/bin/env python3
import argparse
import contextlib
import json
import os
import subprocess
import tempfile
import time

import requests
import ruamel.yaml as yaml
from jira.client import JIRA


def parse_args(argv):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "-s",
        "--splunk_creds",
        help="Creds for Splunk API, user:pass",
        dest="splunk_creds",
        required=True,
    )
    parser.add_argument(
        "-j",
        "--jira_creds",
        help="Creds for JIRA API, user:pass",
        dest="jira_creds",
        required=True,
    )
    parser.add_argument(
        "-n",
        "--no-tick",
        help="Do not create a JIRA ticket",
        action="store_true",
        dest="no_tick",
        default=False,
    )
    parser.add_argument(
        "-m",
        "--manual",
        help="Do not publish a review",
        action="store_true",
        dest="manual_rb",
        default=False,
    )

    parser.add_argument(
        "-f",
        "--file-splunk",
        help="Splunk csv from which to pull data. Defaults to %(default)s",
        dest="file_splunk",
        default="paasta_overprovision_alerts_fired.csv",
    )
    return parser.parse_args(argv)


def tempdir():
    return tempfile.TemporaryDirectory(prefix="repo", dir="/nail/tmp")


@contextlib.contextmanager
def cwd(path):
    pwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(pwd)


@contextlib.contextmanager
def in_tempdir():
    with tempdir() as tmp:
        with cwd(tmp):
            yield


def get_perf_data(creds, filename):
    url = "https://splunk-api.yelpcorp.com/servicesNS/nobody/yelp_performance/search/jobs/export"
    search = (
        "| inputlookup {} |"
        ' eval _time = search_time | where _time > relative_time(now(),"-7d")'
    ).format(filename)
    data = {"output_mode": "json", "search": search}
    creds = creds.split(":")
    resp = requests.post(url, data=data, auth=(creds[0], creds[1]))
    resp_text = resp.text.split("\n")
    resp_text = [x for x in resp_text if x]
    resp_text = [json.loads(x) for x in resp_text]

    services_to_update = []
    for d in resp_text:
        criteria = d["result"]["criteria"].split()
        serv = {}
        serv["service"] = criteria[0]
        serv["cluster"] = criteria[1]
        serv["instance"] = criteria[2]
        serv["cpus"] = d["result"]["suggested_cpus"]
        serv["owner"] = d["result"]["service_owner"]
        serv["money"] = d["result"]["estimated_monthly_savings"]
        serv["date"] = d["result"]["_time"].split(" ")[0]
        serv["old_cpus"] = d["result"]["current_cpus"]
        serv["project"] = d["result"]["project"]
        services_to_update.append(serv)

    return services_to_update


def clone(branch_name):
    remote = "git@sysgit.yelpcorp.com:yelpsoa-configs"
    subprocess.check_call(("git", "clone", remote, "."))
    subprocess.check_call(("git", "checkout", "-b", branch_name))


def commit(filename, serv):
    message = "Updating {} for {}provisioned cpu from {} to {} cpus".format(
        filename, serv["state"], serv["old_cpus"], serv["cpus"]
    )
    subprocess.check_call(("git", "add", filename))
    subprocess.check_call(("git", "commit", "-n", "-m", message))


def get_reviewers_in_group(group_name):
    """Using rbt's target-groups argument overrides our configured default review groups.
    So we'll expand the group into usernames and pass those users in the group individually.
    """
    rightsizer_reviewers = json.loads(
        subprocess.check_output(
            (
                "rbt",
                "api-get",
                "--server",
                "https://reviewboard.yelpcorp.com",
                f"groups/{group_name}/users/",
            )
        ).decode("UTF-8")
    )
    return [user.get("username", "") for user in rightsizer_reviewers.get("users", {})]


def get_reviewers(filename):
    recent_authors = set()
    authors = (
        subprocess.check_output(("git", "log", "--format=%ae", "--", filename))
        .decode("UTF-8")
        .splitlines()
    )

    authors = [x.split("@")[0] for x in authors]
    for author in authors:
        recent_authors.add(author)
        if len(recent_authors) >= 3:
            break
    return recent_authors


def review(filename, summary, description, manual_rb):
    all_reviewers = get_reviewers(filename).union(get_reviewers_in_group("right-sizer"))
    reviewers_arg = " ".join(all_reviewers)
    if manual_rb:
        subprocess.check_call(
            (
                "review-branch",
                f"--summary={summary}",
                f"--description={description}",
                "--reviewers",
                reviewers_arg,
                "--server",
                "https://reviewboard.yelpcorp.com",
            )
        )
    else:
        subprocess.check_call(
            (
                "review-branch",
                f"--summary={summary}",
                f"--description={description}",
                "-p",
                "--reviewers",
                reviewers_arg,
                "--server",
                "https://reviewboard.yelpcorp.com",
            )
        )


def edit_soa_configs(filename, instance, cpu):
    with open(filename, "r") as fi:
        yams = fi.read()
        yams = yams.replace("cpus: .", "cpus: 0.")
        data = yaml.round_trip_load(yams, preserve_quotes=True)

    instdict = data[instance]
    instdict["cpus"] = cpu
    out = yaml.round_trip_dump(data, width=120)

    with open(filename, "w") as fi:
        fi.write(out)


def create_jira_ticket(serv, creds, description):
    creds = creds.split(":")
    options = {"server": "https://jira.yelpcorp.com"}
    jira_cli = JIRA(options=options, basic_auth=(creds[0], creds[1]))
    jira_ticket = {}
    # Sometimes a project has required fields we can't predict
    try:
        jira_ticket = {
            "project": {"key": serv["project"]},
            "description": description,
            "issuetype": {"name": "Improvement"},
            "labels": ["perf-watching", "paasta-rightsizer"],
            "summary": "{s}.{i} in {c} may be {o}provisioned".format(
                s=serv["service"],
                i=serv["instance"],
                c=serv["cluster"],
                o=serv["state"],
            ),
        }
        tick = jira_cli.create_issue(fields=jira_ticket)
    except Exception:
        jira_ticket["project"] = {"key": "PEOBS"}
        jira_ticket["labels"].append(serv["service"])
        tick = jira_cli.create_issue(fields=jira_ticket)
    return tick.key


def _get_dashboard_qs_param(param, value):
    # Some dashboards may ask for query string params like param=value, but not this provider.
    return f"variables%5B%5D={param}%3D{param}:{value}"


def main(argv=None):
    args = parse_args(argv)
    services_to_update = get_perf_data(args.splunk_creds, args.file_splunk)

    for serv in services_to_update:
        filename = "{}/{}.yaml".format(serv["service"], serv["cluster"])
        cpus = float(serv["cpus"])
        provisioned_state = "over"
        if cpus > float(serv["old_cpus"]):
            provisioned_state = "under"

        serv["state"] = provisioned_state
        ticket_desc = (
            "This ticket and CR have been auto-generated to help keep PaaSTA right-sized."
            "\nPEOBS will review this CR and give a shipit. Then an ops deputy from your team can merge"
            " if these values look good for your service after review."
            "\nOpen an issue with any concerns and someone from PEOBS will respond."
            "\nWe suspect that {s}.{i} in {c} may have been {o}-provisioned"
            " during the 1 week prior to {d}. It initially had {x} cpus, but based on the below dashboard,"
            " we recommend {y} cpus."
            "\n- Dashboard: https://y.yelpcorp.com/{o}provisioned?{cluster_param}&{service_param}&{instance_param}"
            "\n- Service owner: {n}"
            "\n- Estimated monthly excess cost: ${m}"
            "\n\nFor more information and sizing examples for larger services:"
            "\n- Runbook: https://y.yelpcorp.com/rb-provisioning-alert"
            "\n- Alert owner: pe-observability@yelp.com"
        ).format(
            s=serv["service"],
            c=serv["cluster"],
            i=serv["instance"],
            o=provisioned_state,
            d=serv["date"],
            n=serv["owner"],
            m=serv["money"],
            x=serv["old_cpus"],
            y=serv["cpus"],
            cluster_param=_get_dashboard_qs_param(
                "paasta_cluster", serv["cluster"].replace("marathon-", "")
            ),
            service_param=_get_dashboard_qs_param("paasta_service", serv["service"]),
            instance_param=_get_dashboard_qs_param("paasta_instance", serv["instance"]),
        )
        summary = f"Rightsizing {serv['service']}.{serv['instance']} in {serv['cluster']} to make it not have {provisioned_state}-provisioned cpu"  # noqa: E501
        branch = ""
        if args.no_tick:
            branch = "rightsize-{}".format(int(time.time()))
        else:
            branch = create_jira_ticket(serv, args.jira_creds, ticket_desc)

        with in_tempdir():
            clone(branch)
            edit_soa_configs(filename, serv["instance"], cpus)
            try:
                commit(filename, serv)
                review(filename, summary, ticket_desc, args.manual_rb)
            except Exception:
                print(
                    (
                        "\nUnable to push changes to {f}. Check if {f} conforms to"
                        "yelpsoa-configs yaml rules. No review created. To see the"
                        "cpu suggestion for this service check {t}."
                    ).format(f=filename, t=branch)
                )
                continue


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import time

import a_sync
import simplejson as json

from paasta_tools import mesos_tools
from paasta_tools.mesos.exceptions import SlaveDoesNotExist
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print


def get_container_info_from_task(task):
    for status in task["statuses"]:
        if status["state"] != "TASK_RUNNING":
            continue
        container_id = (
            status.get("container_status", {}).get("container_id", {}).get("value")
        )
        time_start = status.get("timestamp")
        return container_id, time_start
    return None, None


def get_paasta_service_instance_from_task(task):
    try:
        docker_params = task["container"].get("docker", {}).get("parameters", [])
    except KeyError:
        return None, None
    service, instance = None, None
    for param in docker_params:
        if param["key"] == "label":
            label = param["value"]
            if label.startswith("paasta_service="):
                service = label.split("=")[1]
            if label.startswith("paasta_instance="):
                instance = label.split("=")[1]
    return service, instance


async def get_pool_from_task(task):
    try:
        attributes = (await task.slave())["attributes"]
        return attributes.get("pool", "default")
    except SlaveDoesNotExist:
        return None


@a_sync.to_blocking
async def get_task_allocation_info():
    tasks = await mesos_tools.get_cached_list_of_running_tasks_from_frameworks()
    info_list = []
    for task in tasks:
        info = {"resources": task["resources"]}
        info["mesos_container_id"], info["start_time"] = get_container_info_from_task(
            task
        )
        info["paasta_service"], info[
            "paasta_instance"
        ] = get_paasta_service_instance_from_task(task)
        info["paasta_pool"] = await get_pool_from_task(task)
        info_list.append(info)
    return info_list


def main():
    cluster = load_system_paasta_config().get_cluster()
    info_list = get_task_allocation_info()
    timestamp = time.time()
    for info in info_list:
        info["cluster"] = cluster
        info["timestamp"] = timestamp
        paasta_print(json.dumps(info))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./cleanup_kubernetes_crd.py [options]

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -c <cluster>, --cluster <cluster>: Specify a kubernetes cluster name
- -v, --verbose: Verbose output
- -n, --dry-run: Only report what would have been deleted
"""
import argparse
import logging
import sys

import service_configuration_lib
from kubernetes.client import V1DeleteOptions
from kubernetes.client.rest import ApiException

from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_system_paasta_config

log = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Removes stale kubernetes CRDs.")
    parser.add_argument(
        "-c",
        "--cluster",
        dest="cluster",
        metavar="CLUSTER",
        default=None,
        help="Kubernetes cluster name",
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", dest="verbose", default=False
    )
    parser.add_argument(
        "-n", "--dry-run", action="store_true", dest="dry_run", default=False
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    soa_dir = args.soa_dir
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    if args.cluster:
        cluster = args.cluster
    else:
        system_paasta_config = load_system_paasta_config()
        cluster = system_paasta_config.get_cluster()

    kube_client = KubeClient()

    success = cleanup_kube_crd(
        kube_client=kube_client, cluster=cluster, soa_dir=soa_dir, dry_run=args.dry_run
    )
    sys.exit(0 if success else 1)


def cleanup_kube_crd(
    kube_client: KubeClient,
    cluster: str,
    soa_dir: str = DEFAULT_SOA_DIR,
    dry_run: bool = False,
) -> bool:
    existing_crds = kube_client.apiextensions.list_custom_resource_definition(
        label_selector="yelp.com/paasta_service"
    )

    success = True
    for crd in existing_crds.items:
        service = crd.metadata.labels["yelp.com/paasta_service"]
        if not service:
            log.error(f"CRD {crd.metadata.name} has empty paasta_service label")
            continue

        crd_config = service_configuration_lib.read_extra_service_information(
            service, f"crd-{cluster}", soa_dir=soa_dir
        )
        if crd_config:
            log.debug(f"CRD {crd.metadata.name} declaration found in {service}")
            continue

        log.info(f"CRD {crd.metadata.name} not found in {service} service")
        if dry_run:
            log.info("not deleting in dry-run mode")
            continue

        try:
            kube_client.apiextensions.delete_custom_resource_definition(
                name=crd.metadata.name, body=V1DeleteOptions()
            )
            log.info(f"deleted {crd.metadata.name} for {cluster}:{service}")
        except ApiException as exc:
            log.error(
                f"error deploying crd for {cluster}:{service}, "
                f"status: {exc.status}, reason: {exc.reason}"
            )
            log.debug(exc.body)
            success = False

    return success


if __name__ == "__main__":
    main()



import staticconf

CLUSTERMAN_YAML_FILE_PATH = "/nail/srv/configs/clusterman.yaml"
CLUSTERMAN_METRICS_YAML_FILE_PATH = "/nail/srv/configs/clusterman_metrics.yaml"


def get_clusterman_metrics():
    try:
        import clusterman_metrics
        import clusterman_metrics.util.costs

        clusterman_yaml = CLUSTERMAN_YAML_FILE_PATH
        staticconf.YamlConfiguration(
            CLUSTERMAN_METRICS_YAML_FILE_PATH, namespace="clusterman_metrics"
        )
    except (ImportError, FileNotFoundError):
        # our cluster autoscaler is not currently open source, sorry!
        clusterman_metrics = None
        clusterman_yaml = None

    return clusterman_metrics, clusterman_yaml



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import collections
import csv
import socket
from typing import cast
from typing import Collection
from typing import DefaultDict
from typing import Dict
from typing import Iterable
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar

import requests
from kubernetes.client import V1Node
from kubernetes.client import V1Pod
from mypy_extensions import TypedDict

from paasta_tools import kubernetes_tools
from paasta_tools import marathon_tools
from paasta_tools import mesos_tools
from paasta_tools.mesos.exceptions import NoSlavesAvailableError
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import DeployBlacklist
from paasta_tools.utils import get_user_agent
from paasta_tools.utils import InstanceConfig
from paasta_tools.utils import SystemPaastaConfig


HaproxyBackend = TypedDict(
    "HaproxyBackend",
    {
        "check_code": str,
        "check_duration": str,
        "check_status": str,
        "lastchg": str,
        "pxname": str,
        "svname": str,
        "status": str,
    },
    total=False,
)


def retrieve_haproxy_csv(
    synapse_host: str, synapse_port: int, synapse_haproxy_url_format: str, scope: str
) -> Iterable[Dict[str, str]]:
    """Retrieves the haproxy csv from the haproxy web interface

    :param synapse_host_port: A string in host:port format that this check
                              should contact for replication information.
    :returns reader: a csv.DictReader object
    """
    synapse_uri = synapse_haproxy_url_format.format(
        host=synapse_host, port=synapse_port, scope=scope
    )

    # timeout after 1 second and retry 3 times
    haproxy_request = requests.Session()
    haproxy_request.headers.update({"User-Agent": get_user_agent()})
    haproxy_request.mount("http://", requests.adapters.HTTPAdapter(max_retries=3))
    haproxy_request.mount("https://", requests.adapters.HTTPAdapter(max_retries=3))
    haproxy_response = haproxy_request.get(synapse_uri, timeout=1)
    haproxy_data = haproxy_response.text
    reader = csv.DictReader(haproxy_data.splitlines())
    return reader


def get_backends(
    service: str, synapse_host: str, synapse_port: int, synapse_haproxy_url_format: str
) -> List[HaproxyBackend]:
    """Fetches the CSV from haproxy and returns a list of backends,
    regardless of their state.

    :param service: If None, return backends for all services, otherwise only return backends for this particular
                    service.
    :param synapse_host_port: A string in host:port format that this check
                              should contact for replication information.
    :returns backends: A list of dicts representing the backends of all
                       services or the requested service
    """
    if service:
        services = [service]
    else:
        services = None
    return get_multiple_backends(
        services,
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )


def get_multiple_backends(
    services: Optional[Collection[str]],
    synapse_host: str,
    synapse_port: int,
    synapse_haproxy_url_format: str,
) -> List[HaproxyBackend]:
    """Fetches the CSV from haproxy and returns a list of backends,
    regardless of their state.

    :param services: If None, return backends for all services, otherwise only return backends for these particular
                     services.
    :param synapse_host_port: A string in host:port format that this check
                              should contact for replication information.
    :returns backends: A list of dicts representing the backends of all
                       services or the requested service
    """

    if services is not None and len(services) == 1:
        scope, = services
    else:
        # Maybe if there's like two or three services we could make two queries, or find the longest common substring.
        # For now let's just hope this is rare and fetch all data.
        scope = ""

    reader = retrieve_haproxy_csv(
        synapse_host,
        synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
        scope=scope,
    )
    backends = []

    for line in reader:
        # clean up two irregularities of the CSV output, relative to
        # DictReader's behavior there's a leading "# " for no good reason:
        line["pxname"] = line.pop("# pxname")
        # and there's a trailing comma on every line:
        line.pop("")

        # Look for the service in question and ignore the fictional
        # FRONTEND/BACKEND hosts, use starts_with so that hosts that are UP
        # with 1/X healthchecks to go before going down get counted as UP:
        ha_slave, ha_service = line["svname"], line["pxname"]
        if (services is None or ha_service in services) and ha_slave not in (
            "FRONTEND",
            "BACKEND",
        ):
            backends.append(cast(HaproxyBackend, line))

    return backends


def load_smartstack_info_for_service(
    service: str,
    namespace: str,
    blacklist: DeployBlacklist,
    system_paasta_config: SystemPaastaConfig,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> Dict[str, Dict[str, int]]:
    """Retrieves number of available backends for given services

    :param service_instances: A list of tuples of (service, instance)
    :param namespaces: list of Smartstack namespaces
    :param blacklist: A list of blacklisted location tuples in the form (location, value)
    :param system_paasta_config: A SystemPaastaConfig object representing the system configuration.
    :returns: a dictionary of the form

    ::

        {
          'location_type': {
              'unique_location_name': {
                  'service.instance': <# ofavailable backends>
              },
              'other_unique_location_name': ...
          }
        }

    """
    service_namespace_config = marathon_tools.load_service_namespace_config(
        service=service, namespace=namespace, soa_dir=soa_dir
    )
    discover_location_type = service_namespace_config.get_discover()
    return get_smartstack_replication_for_attribute(
        attribute=discover_location_type,
        service=service,
        namespace=namespace,
        blacklist=blacklist,
        system_paasta_config=system_paasta_config,
    )


def get_smartstack_replication_for_attribute(
    attribute: str,
    service: str,
    namespace: str,
    blacklist: DeployBlacklist,
    system_paasta_config: SystemPaastaConfig,
) -> Dict[str, Dict[str, int]]:
    """Loads smartstack replication from a host with the specified attribute

    :param attribute: a Mesos attribute
    :param service: A service name, like 'example_service'
    :param namespace: A particular smartstack namespace to inspect, like 'main'
    :param constraints: A list of Marathon constraints to restrict which synapse hosts to query
    :param blacklist: A list of blacklisted location tuples in the form of (location, value)
    :param system_paasta_config: A SystemPaastaConfig object representing the system configuration.
    :returns: a dictionary of the form {'<unique_attribute_value>': <smartstack replication hash>}
              (the dictionary will contain keys for unique all attribute values)
    """
    replication_info = {}
    filtered_slaves = mesos_tools.get_all_slaves_for_blacklist_whitelist(
        blacklist=blacklist, whitelist=None
    )
    if not filtered_slaves:
        raise NoSlavesAvailableError

    attribute_slave_dict = mesos_tools.get_mesos_slaves_grouped_by_attribute(
        slaves=filtered_slaves, attribute=attribute
    )

    full_name = compose_job_id(service, namespace)

    for value, hosts in attribute_slave_dict.items():
        # arbitrarily choose the first host with a given attribute to query for replication stats
        synapse_host = hosts[0]["hostname"]
        repl_info = get_replication_for_services(
            synapse_host=synapse_host,
            synapse_port=system_paasta_config.get_synapse_port(),
            synapse_haproxy_url_format=system_paasta_config.get_synapse_haproxy_url_format(),
            services=[full_name],
        )
        replication_info[value] = repl_info

    return replication_info


def get_replication_for_all_services(
    synapse_host: str, synapse_port: int, synapse_haproxy_url_format: str
) -> Dict[str, int]:
    """Returns the replication level for all services known to this synapse haproxy

    :param synapse_host: The host that this check should contact for replication information.
    :param synapse_port: The port number that this check should contact for replication information.
    :param synapse_haproxy_url_format: The format of the synapse haproxy URL.
    :returns available_instance_counts: A dictionary mapping the service names
                                        to an integer number of available replicas.
    """
    backends = get_multiple_backends(
        services=None,
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )
    return collections.Counter([b["pxname"] for b in backends if backend_is_up(b)])


def get_replication_for_services(
    synapse_host: str,
    synapse_port: int,
    synapse_haproxy_url_format: str,
    services: Collection[str],
) -> Dict[str, int]:
    """Returns the replication level for the provided services

    This check is intended to be used with an haproxy load balancer, and
    relies on the implementation details of that choice.

    :param synapse_host: The host that this check should contact for replication information.
    :param synapse_port: The port number that this check should contact for replication information.
    :param synapse_haproxy_url_format: The format of the synapse haproxy URL.
    :param services: A list of strings that are the service names
                          that should be checked for replication.

    :returns available_instance_counts: A dictionary mapping the service names
                                  to an integer number of available
                                  replicas
    :returns None: If it cannot connect to the specified synapse host and port
    """
    backends = get_multiple_backends(
        services=services,
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )

    counter = collections.Counter([b["pxname"] for b in backends if backend_is_up(b)])
    return {sn: counter[sn] for sn in services}


def backend_is_up(backend: HaproxyBackend) -> bool:
    """Returns whether a server is receiving traffic in HAProxy.

    :param backend: backend dict, like one of those returned by smartstack_tools.get_multiple_backends.

    :returns is_up: Whether the backend is in a state that receives traffic.
    """
    return str(backend["status"]).startswith("UP")


def ip_port_hostname_from_svname(svname: str) -> Tuple[str, int, str]:
    """This parses the haproxy svname that smartstack creates.
    In old versions of synapse, this is in the format ip:port_hostname.
    In versions newer than dd5843c987740a5d5ce1c83b12b258b7253784a8 it is
    hostname_ip:port

    :param svname: A svname, in either of the formats described above
    :returns ip_port_hostname: A tuple of ip, port, hostname.
    """
    # split into parts
    parts = set(svname.split("_"))

    # find those that can be split by : - this is the ip:port
    # there will only be 1 of these
    ip_ports = {part for part in parts if len(part.split(":")) == 2}

    # the one *not* in the list is the hostname
    hostname = parts.difference(ip_ports).pop()

    ip, port = ip_ports.pop().split(":")
    return ip, int(port), hostname


def get_registered_marathon_tasks(
    synapse_host: str,
    synapse_port: int,
    synapse_haproxy_url_format: str,
    service: str,
    marathon_tasks: Iterable[marathon_tools.MarathonTask],
) -> List[marathon_tools.MarathonTask]:
    """Returns the marathon tasks that are registered in haproxy under a given service (nerve_ns).

    :param synapse_host: The host that this check should contact for replication information.
    :param synapse_port: The port that this check should contact for replication information.
    :param synapse_haproxy_url_format: The format of the synapse haproxy URL.
    :param service: A list of strings that are the service names that should be checked for replication.
    :param marathon_tasks: A list of MarathonTask objects, whose tasks we will check for in the HAProxy status.
    """
    backends = get_multiple_backends(
        [service],
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )
    healthy_tasks = []
    for backend, task in match_backends_and_tasks(backends, marathon_tasks):
        if (
            backend is not None
            and task is not None
            and backend["status"].startswith("UP")
        ):
            healthy_tasks.append(task)
    return healthy_tasks


def are_services_up_on_ip_port(
    synapse_host: str,
    synapse_port: int,
    synapse_haproxy_url_format: str,
    services: Collection[str],
    host_ip: str,
    host_port: int,
) -> bool:
    backends = get_multiple_backends(
        services,
        synapse_host=synapse_host,
        synapse_port=synapse_port,
        synapse_haproxy_url_format=synapse_haproxy_url_format,
    )
    backends_by_ip_port: DefaultDict[
        Tuple[str, int], List[HaproxyBackend]
    ] = collections.defaultdict(list)

    for backend in backends:
        ip, port, _ = ip_port_hostname_from_svname(backend["svname"])
        backends_by_ip_port[ip, port].append(backend)

    backends_on_ip = backends_by_ip_port[host_ip, host_port]
    # any backend being up is okay because a previous backend
    # may have had the same IP and synapse only removes them completely
    # after some time
    services_with_atleast_one_backend_up = {service: False for service in services}
    for service in services:
        for be in backends_on_ip:
            if be["pxname"] == service and backend_is_up(be):
                services_with_atleast_one_backend_up[service] = True
    return all(services_with_atleast_one_backend_up.values())


def match_backends_and_tasks(
    backends: Iterable[HaproxyBackend], tasks: Iterable[marathon_tools.MarathonTask]
) -> List[Tuple[Optional[HaproxyBackend], Optional[marathon_tools.MarathonTask]]]:
    """Returns tuples of matching (backend, task) pairs, as matched by IP and port. Each backend will be listed exactly
    once, and each task will be listed once per port. If a backend does not match with a task, (backend, None) will
    be included. If a task's port does not match with any backends, (None, task) will be included.

    :param backends: An iterable of haproxy backend dictionaries, e.g. the list returned by
                     smartstack_tools.get_multiple_backends.
    :param tasks: An iterable of MarathonTask objects.
    """

    # { (ip, port) : [backend1, backend2], ... }
    backends_by_ip_port: DefaultDict[
        Tuple[str, int], List[HaproxyBackend]
    ] = collections.defaultdict(list)
    backend_task_pairs = []

    for backend in backends:
        ip, port, _ = ip_port_hostname_from_svname(backend["svname"])
        backends_by_ip_port[ip, port].append(backend)

    for task in tasks:
        ip = socket.gethostbyname(task.host)
        for port in task.ports:
            for backend in backends_by_ip_port.pop((ip, port), [None]):
                backend_task_pairs.append((backend, task))

    # we've been popping in the above loop, so anything left didn't match a marathon task.
    for backends in backends_by_ip_port.values():
        for backend in backends:
            backend_task_pairs.append((backend, None))

    return backend_task_pairs


def match_backends_and_pods(
    backends: Iterable[HaproxyBackend], pods: Iterable[V1Pod]
) -> List[Tuple[Optional[HaproxyBackend], Optional[V1Pod]]]:
    """Returns tuples of matching (backend, pod) pairs, as matched by IP. Each backend will be listed exactly
    once. If a backend does not match with a pod, (backend, None) will be included.
    If a pod's IP does not match with any backends, (None, pod) will be included.

    :param backends: An iterable of haproxy backend dictionaries, e.g. the list returned by
                     smartstack_tools.get_multiple_backends.
    :param pods: An iterable of V1Pod objects.
    """

    # { ip : [backend1, backend2], ... }
    backends_by_ip: DefaultDict[str, List[HaproxyBackend]] = collections.defaultdict(
        list
    )
    backend_pod_pairs = []

    for backend in backends:
        ip, port, _ = ip_port_hostname_from_svname(backend["svname"])
        backends_by_ip[ip].append(backend)

    for pod in pods:
        ip = pod.status.pod_ip
        for backend in backends_by_ip.pop(ip, [None]):
            backend_pod_pairs.append((backend, pod))

    # we've been popping in the above loop, so anything left didn't match a k8s pod.
    for backends in backends_by_ip.values():
        for backend in backends:
            backend_pod_pairs.append((backend, None))

    return backend_pod_pairs


_MesosSlaveDict = TypeVar(
    "_MesosSlaveDict", bound=Dict
)  # no type has been defined in mesos_tools for these yet.


class SmartstackHost(NamedTuple):
    hostname: str
    pool: str


class SmartstackReplicationChecker(abc.ABC):
    """Base class for checking smartstack replication. Extendable for different frameworks.

    Optimized for multiple queries. Gets the list of backends from synapse-`roxy
    only once per location and reuse it in all subsequent calls of
    SmartstackReplicationChecker.get_replication_for_instance().

    get_allowed_locations_and_hosts must be implemented in sub class
    """

    def __init__(self, system_paasta_config: SystemPaastaConfig) -> None:
        self._synapse_port = system_paasta_config.get_synapse_port()
        self._synapse_haproxy_url_format = (
            system_paasta_config.get_synapse_haproxy_url_format()
        )
        self._system_paasta_config = system_paasta_config
        self._cache: Dict[str, Dict[str, int]] = {}

    @abc.abstractmethod
    def get_allowed_locations_and_hosts(
        self, instance_config: InstanceConfig
    ) -> Dict[str, Sequence[SmartstackHost]]:
        pass

    def get_replication_for_instance(
        self, instance_config: InstanceConfig
    ) -> Dict[str, Dict[str, int]]:
        """Returns the number of registered instances in each discoverable location.

        :param instance_config: An instance of MarathonServiceConfig.
        :returns: a dict {'location_type': {'service.instance': int}}
        """
        replication_info = {}
        attribute_host_dict = self.get_allowed_locations_and_hosts(instance_config)
        instance_pool = instance_config.get_pool()
        for location, hosts in attribute_host_dict.items():
            hostname = self.get_first_host_in_pool(hosts, instance_pool)
            replication_info[location] = self._get_replication_info(
                location, hostname, instance_config
            )
        return replication_info

    def get_first_host_in_pool(self, hosts: Sequence[SmartstackHost], pool: str) -> str:
        for host in hosts:
            if host.pool == pool:
                return host.hostname
        return hosts[0].hostname

    def _get_replication_info(
        self, location: str, hostname: str, instance_config: InstanceConfig
    ) -> Dict[str, int]:
        """Returns service.instance and the number of instances registered in smartstack
        at the location as a dict.

        :param location: A string that identifies a habitat, a region and etc.
        :param hostname: A mesos slave hostname to read replication information from.
        :param instance_config: An instance of MarathonServiceConfig.
        :returns: A dict {"service.instance": number_of_instances}.
        """
        full_name = compose_job_id(instance_config.service, instance_config.instance)
        if location not in self._cache:
            self._cache[location] = get_replication_for_all_services(
                synapse_host=hostname,
                synapse_port=self._synapse_port,
                synapse_haproxy_url_format=self._synapse_haproxy_url_format,
            )
        return {full_name: self._cache[location][full_name]}


class MesosSmartstackReplicationChecker(SmartstackReplicationChecker):
    """Retrieves the number of registered instances in each discoverable location.

    Based on SmartstackReplicationChecker takes mesos slaves as an argument to filter
    which services are allowed to run where.
    :Example:

    >>> from paasta_tools.mesos_tools import get_slaves
    >>> from paasta_tools.utils import load_system_paasta_config
    >>> from paasta_tools.marathon_tools import load_marathon_service_config
    >>> from paasta_tools.smartstack_tools import SmartstackReplicationChecker
    >>>
    >>> mesos_slaves = get_slaves()
    >>> system_paasta_config = load_system_paasta_config()
    >>> instance_config = load_marathon_service_config(service='fake_service',
    ...                       instance='fake_instance', cluster='norcal-stagef')
    >>>
    >>> c = SmartstackReplicationChecker(mesos_slaves, system_paasta_config)
    >>> c.get_replication_for_instance(instance_config)
    {'uswest1-stagef': {'fake_service.fake_instance': 2}}
    >>>
    """

    def __init__(
        self,
        mesos_slaves: List[_MesosSlaveDict],
        system_paasta_config: SystemPaastaConfig,
    ) -> None:
        self._mesos_slaves = mesos_slaves
        super().__init__(system_paasta_config=system_paasta_config)

    def get_allowed_locations_and_hosts(
        self, instance_config: InstanceConfig
    ) -> Dict[str, Sequence[SmartstackHost]]:
        """Returns a dict of locations and lists of corresponding mesos slaves
        where deployment of the instance is allowed.

        :param instance_config: An instance of MarathonServiceConfig
        :returns: A dict {"uswest1-prod": [SmartstackHost(), SmartstackHost(), ...]}
        """
        discover_location_type = marathon_tools.load_service_namespace_config(
            service=instance_config.service,
            namespace=instance_config.instance,
            soa_dir=instance_config.soa_dir,
        ).get_discover()
        attribute_to_slaves = mesos_tools.get_mesos_slaves_grouped_by_attribute(
            slaves=self._mesos_slaves, attribute=discover_location_type
        )
        ret: Dict[str, Sequence[SmartstackHost]] = {}
        for attr, slaves in attribute_to_slaves.items():
            ret[attr] = [
                SmartstackHost(
                    hostname=slave["hostname"], pool=slave["attributes"]["pool"]
                )
                for slave in slaves
            ]
        return ret


class KubeSmartstackReplicationChecker(SmartstackReplicationChecker):
    def __init__(
        self, nodes: Sequence[V1Node], system_paasta_config: SystemPaastaConfig
    ) -> None:
        self.nodes = nodes
        super().__init__(system_paasta_config=system_paasta_config)

    def get_allowed_locations_and_hosts(
        self, instance_config: InstanceConfig
    ) -> Dict[str, Sequence[SmartstackHost]]:
        discover_location_type = kubernetes_tools.load_service_namespace_config(
            service=instance_config.service,
            namespace=instance_config.instance,
            soa_dir=instance_config.soa_dir,
        ).get_discover()

        attribute_to_nodes = kubernetes_tools.get_nodes_grouped_by_attribute(
            nodes=self.nodes, attribute=discover_location_type
        )
        ret: Dict[str, Sequence[SmartstackHost]] = {}
        for attr, nodes in attribute_to_nodes.items():
            ret[attr] = [
                SmartstackHost(
                    hostname=node.metadata.labels["yelp.com/hostname"],
                    pool=node.metadata.labels["yelp.com/pool"],
                )
                for node in nodes
            ]
        return ret



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PaaSTA API error handling.
"""
import logging

from pyramid.response import Response
from pyramid.view import view_config


log = logging.getLogger(__name__)


class ApiFailure(Exception):
    def __init__(self, msg, err):
        self.msg = msg
        self.err = err


@view_config(context=ApiFailure)
def api_failure_response(exc, request):
    """Construct an HTTP response with an error status code. This happens when
    the API service has to stop on a 'hard' error. In contrast, the API service
    continues to produce results on a 'soft' error. It will place a 'message'
    field in the output. Multiple 'soft' errors are concatenated in the same
    'message' field when errors happen in the same hierarchy.
    """
    log.error(exc.msg)

    response = Response("ERROR: %s" % exc.msg)
    response.status_int = exc.err
    return response



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Getters for deriving monitoring parameters for mesos-deployed stuff.
This leaves a place for sane defaults that might change depending
on the framework that is asking, and still allows you to set your team
*once* for a service in the general config.

Everything in here is private, and you shouldn't worry about it.
"""
import json
import logging
import os
from typing import Optional

import pysensu_yelp
import service_configuration_lib

from paasta_tools.utils import _log
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import is_under_replicated
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import PaastaNotConfiguredError


log = logging.getLogger(__name__)


def monitoring_defaults(key):
    defaults = {
        "runbook": 'Please set a `runbook` field in your monitoring.yaml. Like "y/rb-mesos". Docs: '
        "https://paasta.readthedocs.io/en/latest/yelpsoa_configs.html#monitoring-yaml",
        "tip": "Please set a `tip` field in your monitoring.yaml. Docs: "
        "https://paasta.readthedocs.io/en/latest/yelpsoa_configs.html#monitoring-yaml",
        "ticket": False,
        "project": None,
        "realert_every": -1,
        "tags": [],
    }
    return defaults.get(key, None)


def get_team(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("team", overrides, service, soa_dir)


def get_runbook(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("runbook", overrides, service, soa_dir)


def get_tip(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("tip", overrides, service, soa_dir)


def get_notification_email(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value(
        "notification_email", overrides, service, soa_dir
    )


def get_page(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("page", overrides, service, soa_dir)


def get_alert_after(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("alert_after", overrides, service, soa_dir)


def get_realert_every(
    overrides, service, soa_dir=DEFAULT_SOA_DIR, monitoring_defaults=monitoring_defaults
):
    return __get_monitoring_config_value(
        "realert_every",
        overrides=overrides,
        service=service,
        soa_dir=soa_dir,
        monitoring_defaults=monitoring_defaults,
    )


def get_check_every(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("check_every", overrides, service, soa_dir)


def get_irc_channels(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("irc_channels", overrides, service, soa_dir)


def get_slack_channels(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("slack_channels", overrides, service, soa_dir)


def get_dependencies(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("dependencies", overrides, service, soa_dir)


def get_ticket(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("ticket", overrides, service, soa_dir)


def get_project(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("project", overrides, service, soa_dir)


def get_priority(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("priority", overrides, service, soa_dir)


def get_tags(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("tags", overrides, service, soa_dir)


def get_component(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("component", overrides, service, soa_dir)


def get_description(overrides, service, soa_dir=DEFAULT_SOA_DIR):
    return __get_monitoring_config_value("description", overrides, service, soa_dir)


def __get_monitoring_config_value(
    key,
    overrides,
    service,
    soa_dir=DEFAULT_SOA_DIR,
    monitoring_defaults=monitoring_defaults,
):
    general_config = service_configuration_lib.read_service_configuration(
        service, soa_dir=soa_dir
    )
    monitor_config = read_monitoring_config(service, soa_dir=soa_dir)
    service_default = general_config.get(key, monitoring_defaults(key))
    service_default = general_config.get("monitoring", {key: service_default}).get(
        key, service_default
    )
    service_default = monitor_config.get(key, service_default)
    return overrides.get(key, service_default)


def get_sensu_team_data(team):
    """Takes a team and returns the dictionary of Sensu configuration
    settings for that team. The data is in this format:
    https://github.com/Yelp/sensu_handlers#teams
    Returns an empty dictionary if there is nothing to return.

    Not all teams specify all the different types of configuration settings.
    for example, a team may not specify a `notification_email`. It is up
    to the caller of this function to handle that case.
    """
    global_team_data = _load_sensu_team_data()["team_data"]
    return global_team_data.get(team, {})


def _load_sensu_team_data():
    try:
        with open("/etc/sensu/team_data.json") as f:
            team_data = json.load(f)
    except IOError:
        log.warning(
            "No Sensu Team data (/etc/sensu/team_data.json) available. Using empty defaults"
        )
        team_data = {}
    return team_data


def send_event(
    service, check_name, overrides, status, output, soa_dir, ttl=None, cluster=None
):
    """Send an event to sensu via pysensu_yelp with the given information.

    :param service: The service name the event is about
    :param check_name: The name of the check as it appears in Sensu
    :param overrides: A dictionary containing overrides for monitoring options
                      (e.g. notification_email, ticket, page)
    :param status: The status to emit for this event
    :param output: The output to emit for this event
    :param soa_dir: The service directory to read monitoring information from
    :param cluster: The cluster name (optional)
    """
    # This function assumes the input is a string like "mumble.main"
    team = get_team(overrides, service, soa_dir)
    if not team:
        return

    system_paasta_config = load_system_paasta_config()
    if cluster is None:
        try:
            cluster = system_paasta_config.get_cluster()
        except PaastaNotConfiguredError:
            cluster = "localhost"

    result_dict = {
        "name": check_name,
        "runbook": overrides.get("runbook", "http://y/paasta-troubleshooting"),
        "status": status,
        "output": output,
        "team": team,
        "page": get_page(overrides, service, soa_dir),
        "tip": get_tip(overrides, service, soa_dir),
        "notification_email": get_notification_email(overrides, service, soa_dir),
        "check_every": overrides.get("check_every", "1m"),
        "realert_every": overrides.get(
            "realert_every", monitoring_defaults("realert_every")
        ),
        "alert_after": overrides.get("alert_after", "5m"),
        "irc_channels": get_irc_channels(overrides, service, soa_dir),
        "slack_channels": get_slack_channels(overrides, service, soa_dir),
        "ticket": get_ticket(overrides, service, soa_dir),
        "project": get_project(overrides, service, soa_dir),
        "priority": get_priority(overrides, service, soa_dir),
        "source": "paasta-%s" % cluster,
        "tags": get_tags(overrides, service, soa_dir),
        "ttl": ttl,
        "sensu_host": system_paasta_config.get_sensu_host(),
        "sensu_port": system_paasta_config.get_sensu_port(),
        "component": get_component(overrides, service, soa_dir),
        "description": get_description(overrides, service, soa_dir),
    }

    if result_dict.get("sensu_host"):
        pysensu_yelp.send_event(**result_dict)


def read_monitoring_config(service, soa_dir=DEFAULT_SOA_DIR):
    """Read a service's monitoring.yaml file.

    :param service: The service name
    :param soa_dir: THe SOA configuration directory to read from
    :returns: A dictionary of whatever was in soa_dir/name/monitoring.yaml"""
    rootdir = os.path.abspath(soa_dir)
    monitoring_file = os.path.join(rootdir, service, "monitoring.yaml")
    monitor_conf = service_configuration_lib.read_monitoring(monitoring_file)
    return monitor_conf


def list_teams(**kwargs):
    """Loads team data from the system. Returns a set of team names (or empty
    set).
    """
    team_data = _load_sensu_team_data()
    teams = set(team_data.get("team_data", {}).keys())
    return teams


def send_replication_event(instance_config, status, output):
    """Send an event to sensu via pysensu_yelp with the given information.

    :param instance_config: an instance of LongRunningServiceConfig
    :param status: The status to emit for this event
    :param output: The output to emit for this event"""
    # This function assumes the input is a string like "mumble.main"
    monitoring_overrides = instance_config.get_monitoring()
    if "alert_after" not in monitoring_overrides:
        monitoring_overrides["alert_after"] = "2m"
    monitoring_overrides["check_every"] = "1m"
    monitoring_overrides["runbook"] = get_runbook(
        monitoring_overrides, instance_config.service, soa_dir=instance_config.soa_dir
    )

    check_name = "check_paasta_services_replication.%s" % instance_config.job_id
    send_event(
        service=instance_config.service,
        check_name=check_name,
        overrides=monitoring_overrides,
        status=status,
        output=output,
        soa_dir=instance_config.soa_dir,
        cluster=instance_config.cluster,
    )
    _log(
        service=instance_config.service,
        line="Replication: %s" % output,
        component="monitoring",
        level="debug",
        cluster=instance_config.cluster,
        instance=instance_config.instance,
    )


def check_smartstack_replication_for_instance(
    instance_config, expected_count, smartstack_replication_checker
):
    """Check a set of namespaces to see if their number of available backends is too low,
    emitting events to Sensu based on the fraction available and the thresholds defined in
    the corresponding yelpsoa config.

    :param instance_config: an instance of MarathonServiceConfig
    :param smartstack_replication_checker: an instance of SmartstackReplicationChecker
    """

    crit_threshold = instance_config.get_replication_crit_percentage()

    log.info("Checking instance %s in smartstack", instance_config.job_id)
    smartstack_replication_info = smartstack_replication_checker.get_replication_for_instance(
        instance_config
    )

    log.debug(
        "Got smartstack replication info for %s: %s"
        % (instance_config.job_id, smartstack_replication_info)
    )

    if len(smartstack_replication_info) == 0:
        status = pysensu_yelp.Status.CRITICAL
        output = (
            "Service %s has no Smartstack replication info. Make sure the discover key in your smartstack.yaml "
            "is valid!\n"
        ) % instance_config.job_id
        log.error(output)
    else:
        expected_count_per_location = int(
            expected_count / len(smartstack_replication_info)
        )
        output = ""
        output_critical = ""
        output_ok = ""
        under_replication_per_location = []

        for location, available_backends in sorted(smartstack_replication_info.items()):
            num_available_in_location = available_backends.get(
                instance_config.job_id, 0
            )
            under_replicated, ratio = is_under_replicated(
                num_available_in_location, expected_count_per_location, crit_threshold
            )
            if under_replicated:
                output_critical += (
                    "- Service %s has %d out of %d expected instances in %s (CRITICAL: %d%%)\n"
                    % (
                        instance_config.job_id,
                        num_available_in_location,
                        expected_count_per_location,
                        location,
                        ratio,
                    )
                )
            else:
                output_ok += (
                    "- Service %s has %d out of %d expected instances in %s (OK: %d%%)\n"
                    % (
                        instance_config.job_id,
                        num_available_in_location,
                        expected_count_per_location,
                        location,
                        ratio,
                    )
                )
            under_replication_per_location.append(under_replicated)

        output += output_critical
        if output_critical and output_ok:
            output += "\n\n"
            output += "The following locations are OK:\n"
        output += output_ok

        if any(under_replication_per_location):
            status = pysensu_yelp.Status.CRITICAL
            output += (
                "\n\n"
                "What this alert means:\n"
                "\n"
                "  This replication alert means that a SmartStack powered loadbalancer (haproxy)\n"
                "  doesn't have enough healthy backends. Not having enough healthy backends\n"
                "  means that clients of that service will get 503s (http) or connection refused\n"
                "  (tcp) when trying to connect to it.\n"
                "\n"
                "Reasons this might be happening:\n"
                "\n"
                "  The service may simply not have enough copies or it could simply be\n"
                "  unhealthy in that location. There also may not be enough resources\n"
                "  in the cluster to support the requested instance count.\n"
                "\n"
                "Things you can do:\n"
                "\n"
                "  * You can view the logs for the job with:\n"
                "      paasta logs -s %(service)s -i %(instance)s -c %(cluster)s\n"
                "\n"
                "  * Fix the cause of the unhealthy service. Try running:\n"
                "\n"
                "      paasta status -s %(service)s -i %(instance)s -c %(cluster)s -vv\n"
                "\n"
                "  * Widen SmartStack discovery settings\n"
                "  * Increase the instance count\n"
                "\n"
            ) % {
                "service": instance_config.service,
                "instance": instance_config.instance,
                "cluster": instance_config.cluster,
            }
            log.error(output)
        else:
            status = pysensu_yelp.Status.OK
            log.info(output)
    send_replication_event(
        instance_config=instance_config, status=status, output=output
    )


def send_replication_event_if_under_replication(
    instance_config,
    expected_count: int,
    num_available: int,
    sub_component: Optional[str] = None,
):
    crit_threshold = instance_config.get_replication_crit_percentage()
    if sub_component is not None:
        output = (
            "Service %s has %d out of %d expected instances of %s available!\n"
            + "(threshold: %d%%)"
        ) % (
            instance_config.job_id,
            num_available,
            expected_count,
            sub_component,
            crit_threshold,
        )
    else:
        output = (
            "Service %s has %d out of %d expected instances available!\n"
            + "(threshold: %d%%)"
        ) % (instance_config.job_id, num_available, expected_count, crit_threshold)
    under_replicated, _ = is_under_replicated(
        num_available, expected_count, crit_threshold
    )
    if under_replicated:
        output += (
            "\n\n"
            "What this alert means:\n"
            "\n"
            "  This replication alert means that the service PaaSTA can't keep the\n"
            "  requested number of copies up and healthy in the cluster.\n"
            "\n"
            "Reasons this might be happening:\n"
            "\n"
            "  The service may simply be unhealthy. There also may not be enough resources\n"
            "  in the cluster to support the requested instance count.\n"
            "\n"
            "Things you can do:\n"
            "\n"
            "  * Increase the instance count\n"
            "  * Fix the cause of the unhealthy service. Try running:\n"
            "\n"
            "      paasta status -s %(service)s -i %(instance)s -c %(cluster)s -vv\n"
        ) % {
            "service": instance_config.service,
            "instance": instance_config.instance,
            "cluster": instance_config.cluster,
        }
        log.error(output)
        status = pysensu_yelp.Status.CRITICAL
    else:
        log.info(output)
        status = pysensu_yelp.Status.OK
    send_replication_event(
        instance_config=instance_config, status=status, output=output
    )



import errno
import fcntl
import os.path
import random

MAC_ADDRESS_PREFIX = ("02", "52")


class MacAddressException(Exception):
    pass


def reserve_unique_mac_address(lock_directory):
    """ Pick and reserve a unique mac address for a container
    returns (mac_address, lockfile)
    where the mac address is a string in the form of 00:00:00:00:00:00
    and lockfile is a file object that holds an exclusive lock
    """
    for x in range(100):
        random_hex = "{:08x}".format(random.getrandbits(32))
        mac_address = ":".join(
            MAC_ADDRESS_PREFIX
            + (random_hex[0:2], random_hex[2:4], random_hex[4:6], random_hex[6:8])
        )

        lock_filepath = os.path.join(lock_directory, mac_address)
        lock_file = obtain_lock(lock_filepath)
        if lock_file is not None:
            return (mac_address, lock_file)

    raise MacAddressException("Unable to pick unique MAC address")


def obtain_lock(lock_filepath):
    """ Open and obtain a flock on the parameter. Returns a file if successful, None if not
    """
    lock_file = open(lock_filepath, "w")
    try:
        fcntl.flock(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
        return lock_file
    except IOError as err:
        if err.errno != errno.EAGAIN:
            raise
        lock_file.close()
        return None



import logging
import threading
import time

from kazoo.client import KazooState
from kazoo.exceptions import ConnectionClosedError
from kazoo.recipe.election import Election

from paasta_tools.deployd.common import PaastaThread


class PaastaLeaderElection(Election):
    def __init__(self, client, *args, **kwargs):
        self.client = client
        self.control = kwargs.pop("control")
        super().__init__(self.client, *args, **kwargs)
        self.client.add_listener(self.connection_listener)
        self.waiting_for_reconnect = False

    @property
    def log(self):
        name = ".".join([__name__, type(self).__name__])
        return logging.getLogger(name)

    def run(self, func, *args, **kwargs):
        try:
            super().run(func, *args, **kwargs)
        except ConnectionClosedError:
            self.log.error("Zookeeper connection closed so can't tidy up!")
            return

    def connection_listener(self, state):
        self.log.warning(f"Zookeeper connection transitioned to: {state}")
        if state == KazooState.SUSPENDED:
            self.log.warning(
                "Zookeeper connection suspended, waiting to see if it recovers."
            )
            if not self.waiting_for_reconnect:
                self.waiting_for_reconnect = True
                reconnection_checker = PaastaThread(target=self.reconnection_listener)
                reconnection_checker.daemon = True
                reconnection_checker.start()
        elif state == KazooState.LOST:
            self.log.error("Leadership lost, quitting!")
            self._terminate()

    def reconnection_listener(self):
        attempts = 0
        while attempts < 5:
            if self.client.state == KazooState.CONNECTED:
                self.log.warning("Zookeeper connection recovered!")
                self.waiting_for_reconnect = False
                return
            self.log.warning("Waiting for zookeeper connection to recover")
            time.sleep(5)
            attempts += 1
        self.log.error("Connection did not recover, abdicating!")
        self._terminate()

    def _terminate(self):
        thread_info = [
            {"alive": t.is_alive(), "daemon": t.daemon, "name": t.__class__.__name__}
            for t in threading.enumerate()
        ]
        self.log.info(f"Thread info: {thread_info}")
        self.control.put("ABORT")



import json
import os
from distutils.dir_util import copy_tree

from paasta_tools.cli.utils import pick_random_port


def main():
    print("-------------------------------------------------------")
    print(
        "Please run export PAASTA_SYSTEM_CONFIG_DIR=etc_paasta_for_development to continue"
    )
    print(
        "Please set environment variable PAASTA_TEST_CLUSTER to the cluster you want to use."
    )
    print("This is necessary for tron jobs")
    print("-------------------------------------------------------")
    cluster = os.environ.get("PAASTA_TEST_CLUSTER", "norcal-devc")
    config_path = "etc_paasta_for_development"

    copy_tree("/etc/paasta", os.path.join(os.getcwd(), config_path))
    # Generate tron.json
    tron_config = {"tron": {"url": f"http://tron-{cluster}:8089"}}
    with open(config_path + "/tron.json", "w") as f:
        json.dump(tron_config, f)
    # find unused port
    port = pick_random_port("paasta-dev-api")
    # Generate api endpoints
    api_endpoints = {"api_endpoints": {cluster: f"http://localhost:{port}"}}
    api_endpoints_path = os.path.join(os.getcwd(), config_path, "api_endpoints.json")
    os.chmod(api_endpoints_path, 0o777)
    with open(api_endpoints_path, "w") as f:
        json.dump(api_endpoints, f)

    # export config path
    os.environ["PAASTA_SYSTEM_CONFIG_DIR"] = config_path
    os.execl(
        ".tox/py36-linux/bin/python",
        ".tox/py36-linux/bin/python",
        "-m",
        "paasta_tools.api.api",
        *["-D", "-c", cluster, str(port)],
    )


if __name__ == "__main__":
    main()
